#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>

typedef unsigned char byte;

int debug = 0;

uint32_t le32(byte *p)
{
    return *(uint32_t*)p;
}

uint32_t s0, s1, s2, s3, s4, s5, s6, s7;
uint32_t a0, a1, a2, a3, a4, a5, a6, a7;
uint32_t t0, t1, t2, t3, t4, t5, t6, t7;
byte key[16];
byte input[16], output[16];
byte prev[16]; // previous encrypted block

#define MIX(a,b) (((a) & 0xadaaa9aa) | ((b) & 0x52555655))
#define ROTL(a,n) (((a) << (n)) | ((a) >> (32-(n))))
#define ROTR(a,n) (((a) >> (n)) | ((a) << (32-(n))))

void decrypt_block0(void)
{
    byte blkkey[16];
    uint32_t k0123, k4567, k89ab, kcdef, m1, m2, m3, m4, r1, r2, r3, r4;
    byte b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, ba, bb, bc, bd, be, bf;
    int i;

    memcpy(blkkey, key, 16);

    k0123 = le32(&blkkey[64-64]);
    k89ab = le32(&blkkey[72-64]);
    k4567 = le32(&blkkey[68-64]);  // lw  a3,68(sp)
    kcdef = le32(&blkkey[76-64]);  // lw  t5,76(sp)

    m1 = MIX(kcdef, k4567);
    r1 = ROTL(m1, 9);

    b0 = (r1 >> 0x18) + 65;
    b1 = b0 + (r1 >> 0x10) + 72;
    b2 = b1 + (r1 >> 8) + 79;
    b3 = b2 + r1 + 86;

    m2 = MIX(k89ab, k0123);
    r2 = ROTL(m2, 13);

    b4 = b3 + (r2 >> 0x18) + 93;
    b5 = b4 + (r2 >> 0x10) + 100;
    b6 = b5 + (r2 >> 8) + 107;
    b7 = b6 + r2 + 114;

    m3 = MIX(k4567, kcdef);
    r3 = ROTL(m3, 15);

    b8 = b7 + (r3 >> 0x18) + 121;
    b9 = b8 + (r3 >> 0x10) + 128;
    ba = b9 + (r3 >> 8) + 135;
    bb = ba + r3 + 142;

    m4 = MIX(k0123, k89ab);
    r4 = ROTL(m4, 19);

    bc = bb + (r4 >> 0x18) + 149;
    bd = bc + (r4 >> 0x10) + 156;
    be = bd + (r4 >> 8) + 163;
    bf = be + r4 + 170;

    blkkey[79-64] = b0;
    blkkey[78-64] = b1;
    blkkey[77-64] = b2;
    blkkey[76-64] = b3;
    blkkey[75-64] = b4;
    blkkey[74-64] = b5;
    blkkey[73-64] = b6;
    blkkey[72-64] = b7;
    blkkey[71-64] = b8;
    blkkey[70-64] = b9;
    blkkey[69-64] = ba;
    blkkey[68-64] = bb;
    blkkey[67-64] = bc;
    blkkey[66-64] = bd;
    blkkey[65-64] = be;
    blkkey[64-64] = bf;

    for (i = 0; i < 16; i++)
        output[i] = input[i] ^ blkkey[i];

}

#define BYTES(a,b,c,d) (((byte)(a) << 24) | ((byte)(b) << 16) | ((byte)(c) << 8) | (byte)(d))
#define HA(a) (((a) & 0xadaaa9aa))
#define H5(a) (((a) & 0x52555655))

void recover_key(void)
{
    byte blkkey[16];
    uint32_t k0123, k4567, k89ab, kcdef, m1, m2, m3, m4, r1, r2, r3, r4;
    byte b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, ba, bb, bc, bd, be, bf;
    int i;

    // input is the ciphertext of the first block
    // output is the known plaintext (guessed)
    // prev is the IV

    for (i = 0; i < 16; i++)
        blkkey[i] = output[i] ^ input[i] ^ prev[15-i];

    b0 = blkkey[79-64];
    b1 = blkkey[78-64];
    b2 = blkkey[77-64];
    b3 = blkkey[76-64];
    b4 = blkkey[75-64];
    b5 = blkkey[74-64];
    b6 = blkkey[73-64];
    b7 = blkkey[72-64];
    b8 = blkkey[71-64];
    b9 = blkkey[70-64];
    ba = blkkey[69-64];
    bb = blkkey[68-64];
    bc = blkkey[67-64];
    bd = blkkey[66-64];
    be = blkkey[65-64];
    bf = blkkey[64-64];

    k0123 = k4567 = k89ab = kcdef = 0;

    r1 = BYTES(b0-65, b1-b0-72, b2-b1-79, b3-b2-86);
    m1 = ROTR(r1, 9);
    kcdef |= HA(m1);
    k4567 |= H5(m1);

    r2 = BYTES(b4-b3-93, b5-b4-100, b6-b5-107, b7-b6-114);
    m2 = ROTR(r2, 13);
    k89ab |= HA(m2);
    k0123 |= H5(m2);

    r3 = BYTES(b8-b7-121, b9-b8-128, ba-b9-135, bb-ba-142);
    m3 = ROTR(r3, 15);
    k4567 |= HA(m3);
    kcdef |= H5(m3);

    r4 = BYTES(bc-bb-149, bd-bc-156, be-bd-163, bf-be-170);
    m4 = ROTR(r4, 19);
    k0123 |= HA(m4);
    k89ab |= H5(m4);

    *(uint32_t*)&key[0] = k0123;
    *(uint32_t*)&key[4] = k4567;
    *(uint32_t*)&key[8] = k89ab;
    *(uint32_t*)&key[12] = kcdef;
}


void decrypt_block(uint32_t blkno)
{
    byte blkkey[16];
    int i;

    memcpy(blkkey, key, 16);

        a3 = le32(&blkkey[68-64]);  // lw  a3,68(sp)
        t5 = le32(&blkkey[76-64]);  // lw  t5,76(sp)

       //lui a5,0xadaab
       //lui t0,0x52555
        s2 = 0xadaaa9aa;        // addi    s2,a5,-1622 # adaaa9aa
        t0 = 0x52555655;        // addi    t0,t0,1621  # 52555655
        a4 = blkno + 23;        // addi    a4,s1,23
        a2 = t5 & s2;           // and a2,t5,s2
        a5 = a3 & t0;           // and a5,a3,t0
        a4 &= 31;               // andi    a4,a4,31
        a5 |= a2;               // or  a5,a5,a2
        a2 = -a4;               // neg a2,a4
        a4 = a5 >> a4;          // srl a4,a5,a4
        a5 = a5 << (a2 & 31);          // sll a5,a5,a2
        a4 |= a5;               // or  a4,a4,a5
        t1 = a4 >> 0x18;        // srli    t1,a4,0x18
        t1 += 65;               // addi    t1,t1,65
        a5 = a4 >> 0x10;        // srli    a5,a4,0x10
        a2 = t1 & 255;          // andi    a2,t1,255
        a5 = a5 & 255;
        a2 = a2 + a5;
        t4 = le32(&blkkey[64-64]);
        t2 = le32(&blkkey[72-64]);

        a2 = a2 + (72);
        a0 = a4 >> 0x8;
        a1 = a2 & 255;
        a0 = a0 & 255;
        a5 = blkno + (19);
        a1 = a1 + a0;
        a6 = t4 & t0;
        a0 = t2 & s2;
        a5 = a5 & (31);
        a1 = a1 + (79);
        a6 = a6 | a0;
        a7 = -a5;
        a0 = a1 & 255;
        a4 = a4 & 255;
        a5 = a6 >> (a5 & 31);
        a4 = a0 + a4;
        a6 = a6 << (a7 & 31);
        a5 = a5 | a6;
        a0 = a4 + (86);
        a6 = a0 & 255;
        a4 = a5 >> 0x18;
        a6 = a6 + a4;
        a6 = a6 + (93);
        a4 = a5 >> 0x10;
        a7 = a6 & 255;
        a4 = a4 & 255;
        a7 = a7 + a4;
        a7 = a7 + (100);
        t6 = a5 >> 0x8;
        t3 = a7 & 255;
        t6 = t6 & 255;
        a4 = blkno + (17);
        t3 = t3 + t6;
        t5 = t5 & t0;
        a4 = a4 & (31);
        t3 = t3 + (107);
        a3 = a3 & s2;
        a3 = a3 | t5;
        t6 = -a4;
        t5 = t3 & 255;
        a5 = a5 & 255;
        a5 = t5 + a5;
        a4 = a3 >> (a4 & 31);
        a3 = a3 << (t6 & 31);
        a4 = a4 | a3;
        a3 = a5 + (114);
        t5 = a3 & 255;
        a5 = a4 >> 0x18;
        t5 = t5 + a5;
        t5 = t5 + (121);
        a5 = a4 >> 0x10;
        t6 = t5 & 255;
        a5 = a5 & 255;
        t6 = t6 + a5;
        t6 = t6 + (128);
        a5 = a4 >> 0x8;
        s4 = t6 & 255;
        a5 = a5 & 255;
        s4 = s4 + a5;
        s1 = blkno + (13);
        a5 = t4 & s2;
        t0 = t2 & t0;
        s1 = s1 & (31);
        s4 = s4 + (135);
        a5 = a5 | t0;
        t4 = -s1;
        t0 = s4 & 255;
        a4 = a4 & 255;
        s1 = a5 >> (s1 & 31);
        a4 = t0 + a4;
        a5 = a5 << (t4 & 31);
        a5 = s1 | a5;
        t0 = a4 + (142);
        t2 = t0 & 255;
        a4 = a5 >> 0x18;
        t2 = t2 + a4;
        t2 = t2 + (149);
        a4 = a5 >> 0x10;
        s1 = t2 & 255;
        a4 = a4 & 255;
        s1 = s1 + a4;
        s1 = s1 + (156);
        a4 = a5 >> 0x8;
        s2 = s1 & 255;
        a4 = a4 & 255;
        s2 = s2 + a4;
        s2 = s2 + (163);
        a4 = s2 & 255;
        a5 = a5 & 255;
        a5 = a5 + a4;
        a5 = a5 + (170);

        //s6 = s0 + (8);
        //a4 = s3 + (8);
        //t4 = s3 + (12);
        //s5 = s0 + (12);

       //sltu    t4,s6,t4
       //sltu    s5,a4,s5


        s7 = a3 & 255;
        t1 = t1 & 255;
        a2 = a2 & 255;
        a1 = a1 & 255;
        a0 = a0 & 255;
        a6 = a6 & 255;
        a7 = a7 & 255;
        t3 = t3 & 255;
        t5 = t5 & 255;
        t6 = t6 & 255;
        s4 = s4 & 255;
        t0 = t0 & 255;
        t2 = t2 & 255;
        s1 = s1 & 255;
        s2 = s2 & 255;
        a5 = a5 & 255;

   // ; is this an opaque predicate?

      // xori    t4,t4,1
      // xori    a3,s5,1

//   ; mask from   (a5, s2, s1, t2,
//                  t0, s4, t6, t5,
//                  s7, t3, a7, a6,
//                  a0, a1, a2, t1)

        blkkey[79-64] = t1;
        blkkey[78-64] = a2;
        blkkey[77-64] = a1;
        blkkey[76-64] = a0;
        blkkey[75-64] = a6;
        blkkey[74-64] = a7;
        blkkey[73-64] = t3;
        blkkey[72-64] = s7;
        blkkey[71-64] = t5;
        blkkey[70-64] = t6;
        blkkey[69-64] = s4;
        blkkey[68-64] = t0;
        blkkey[67-64] = t2;
        blkkey[66-64] = s1;
        blkkey[65-64] = s2;
        blkkey[64-64] = a5;

    // ; alignment check? opaque predicate leading to equivalent code?

    //    t4 = t4 | a3;
    //   beqz    t4,11bf8   ; if t4 == 0 then xor with mask in regs

    //    a4 = a4 | s6;
    //    a4 = a4 & (3);
    //   bnez    a4,11bf8   ; if a4 != 0 then xor with mask in regs

    for (i = 0; i < 16; i++)
        output[i] = input[i] ^ blkkey[i];

//   ; xor 16 bytes from s0[8:24] to s3[8:24] with mask at 64(sp) 
//
//       lw  a4,8(s0)
//       lw  a5,64(sp)
//       xor a5,a5,a4
//       sw  a5,8(s3)
//       lw  a4,12(s0)
//       lw  a5,68(sp)
//       xor a5,a5,a4
//       sw  a5,12(s3)
//       lw  a4,16(s0)
//       lw  a5,72(sp)
//       xor a5,a5,a4
//       sw  a5,16(s3)
//       lw  a4,20(s0)
//       lw  a5,76(sp)
//       xor a5,a5,a4
//       sw  a5,20(s3)

//status_0_epilogue:
//       sw  zero,0(s3)

}

void dump(byte buf[], int len)
{
    int i;

    for (i = 0; i < len; i++)
        printf("%02x", buf[i]);

    printf("  ");

    for (i = 0; i < len; i++)
        printf("%c", buf[i]);

    printf("\n");
}

int test_schedule(void)
{
    uint32_t i;

    //memcpy(key, "\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\xaa\xbb\xcc\xdd\xee\xff", 16);
    memset(key, 0, 16);
    memset(input, 0, 16);

    for (i = 0; i < 100; i++) {
        decrypt_block(i);
        dump(output, 16);
    }

    return 0;
}

int test(int argc, char *argv[])
{
    FILE *in = fopen(argv[1], "rb");
    //FILE *out = fopen(argv[2], "wb");
    byte iv[16];
    int blkno = 0;
    int i;

    memcpy(key, "\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\xaa\xbb\xcc\xdd\xee\xff", 16);

    fseek(in, 0x70, SEEK_SET); // skip encrypted hmac

    fread(iv, 1, 16, in); // read IV

    // attempt decrypt in CBC mode

    memcpy(prev, iv, 16);

    while (1) {
        if (fread(input, 1, 16, in) < 16)
            break;

        if (0 == blkno)
            decrypt_block0();
        else
            decrypt_block(blkno);

        for (i = 0; i < 16; i++)
            output[i] ^= prev[15-i];

        if (0 == blkno && debug) {
            recover_key();
            dump(key, 16);
            if (0 == memcmp(key, "\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\xaa\xbb\xcc\xdd\xee\xff", 16))
                printf("Good recovery\n");
            else
                printf("Bad recovery\n");            
        }

        memcpy(prev, input, 16);

        dump(output, 16);

        blkno++;
    }

    fclose(in);
    //fclose(out);
    return 0;
}

int good_padding(byte buf[])
{
    int i, n = buf[-1];

    for (i = 0; i < n; i++) {
        if (buf[-1-i] != n)
            return 0;
    }

    return 1;
}

int main(int argc, char *argv[])
{
    FILE *in = fopen(argv[1], "rb");
    char outname[1000], suffix[20];
    byte outbuf[40000], *pobuf;
    FILE *out;
    byte iv[16];
    byte known[14] = "\x5d\x00\x00\x80\x00\xff\xff\xff\xff\xff\xff\xff\xff\x00";
    int b14, b15;
    int blkno;
    int i;

    // brute-force two bytes

    for (b14 = 0; b14 < 256; b14++)
    for (b15 = 0; b15 < 256; b15++)
    {

        blkno = 0;

        pobuf = outbuf;

        fseek(in, 0x70, SEEK_SET); // skip encrypted hmac

        fread(iv, 1, 16, in); // read IV

        memcpy(prev, iv, 16);

        while (1) {
            if (fread(input, 1, 16, in) < 16)
                break;

            if (0 == blkno) {
                memcpy(output, known, 14);
                output[14] = b14;
                output[15] = b15;
                recover_key();
            }
            else {
                decrypt_block(blkno);

                for (i = 0; i < 16; i++)
                    output[i] ^= prev[15-i];
            }

            memcpy(prev, input, 16);

            memcpy(pobuf, output, 16);
            pobuf += 16;

            blkno++;
        }

        // check padding

        if (good_padding(pobuf)) {

            dump(key, 16);

            strcpy(outname, argv[1]);
            sprintf(suffix, ".%02x%02x.lzma", b14, b15);
            strcat(outname, suffix);

            out = fopen(outname, "wb");
            fwrite(outbuf, 1, pobuf-outbuf, out);
            fclose(out);
        }
    }

    fclose(in);

    return 0;
}

