#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <stdint.h>

//extern uint32_t __fastcall xxe(uint8_t cl, uint32_t edx);

uint32_t xxe(uint8_t cl, uint32_t v)
{
    return (v << 4) ^ (v >> 5);
}

uint32_t delta = 0xd2d413b3; // 2 + data[0x1f4b] // checked statically

uint32_t key[4] = {
    0x9AACC69B, // checksum of first 8 dwords pushed on VM 1 stack
    0x89A31D8A, // checksum of next 8 dwords
    0xa5b2cd1c, // not 0xA5B3CD5C, // ROP chain produced!!
    0xB0A5CE54,
};

uint8_t input[32];
uint8_t output[32];

void encrypt(void)
{
    int i, j, q, r;
    uint32_t x3, x4, x5;

    for (q = 0; q < 4; q++) {

        x3 = *(uint32_t *)(&input[8 * q]);
        x4 = *(uint32_t *)(&input[8 * q + 4]);
        x5 = 0;

        for (r = 0; r < 64; r++) {
            x3 += (xxe(4, x4) + x4) ^ (key[x5 & 3] + x5);
            x5 += delta;
            x4 += (xxe(4, x3) + x3) ^ (key[(x5 >> 11) & 3] + x5);
        }
        *(uint32_t *)(&output[8 * q]) = x3;
        *(uint32_t *)(&output[8 * q + 4]) = x4;
    }
}

void decrypt(void)
{
    int i, j, q, r;
    uint32_t x3, x4, x5;

    for (q = 0; q < 4; q++) {

        x3 = *(uint32_t *)(&input[8 * q]);
        x4 = *(uint32_t *)(&input[8 * q + 4]);
        x5 = 64 * delta;

        for (r = 0; r < 64; r++) {
            x4 -= (xxe(4, x3) + x3) ^ (key[(x5 >> 11) & 3] + x5);
            x5 -= delta;
            x3 -= (xxe(4, x4) + x4) ^ (key[x5 & 3] + x5);
        }
        *(uint32_t *)(&output[8 * q]) = x3;
        *(uint32_t *)(&output[8 * q + 4]) = x4;
    }
}

uint64_t bswap(uint64_t x)
{
    uint8_t xx[8], yy[8];

    *(uint64_t *)&xx = x;

    yy[0] = xx[7];
    yy[1] = xx[6];
    yy[2] = xx[5];
    yy[3] = xx[4];
    yy[4] = xx[3];
    yy[5] = xx[2];
    yy[6] = xx[1];
    yy[7] = xx[0];

    return *(uint64_t *)&yy;
}

uint64_t rol(uint64_t x, int n)
{
    if (0 == n)
        return x;

    return (x << n) | (x >> (64-n));
}

uint64_t ror(uint64_t x, int n)
{
    if (0 == n)
        return x;

    return (x >> n) | (x << (64-n));
}

void ROP()
{
    uint64_t rcx, rdx, r8, tmp;

    rdx = 0x756AF83DE1FFE263; // extracted from decrypted ROP chain
    rcx = 0xFA34AE1002547B89;

    rdx = ~rdx;
    rcx = bswap(rcx);

    r8 = *(uint64_t *)&input[0];
    r8 ^= rcx;
    r8 += rdx;
    *(uint64_t *)&output[0] = r8;

    r8 = *(uint64_t *)&input[8];
    r8 ^= rdx;
    r8 = bswap(r8);
    r8 -= rcx;
    *(uint64_t *)&output[8] = r8;

    r8 = *(uint64_t *)&input[16];
    r8 = bswap(r8);
    rcx = rol(rcx, 13);
    r8 = ror(r8, rcx & 63);
    *(uint64_t *)&output[16] = r8;

    r8 = *(uint64_t *)&input[24];
    r8 ^= rcx;
    rdx = ror(rdx, 21);
    tmp = rcx;
    rcx = rdx;
    rdx = tmp;
    r8 = ror(r8, rcx & 63);
    r8 += rdx;
    *(uint64_t *)&output[24] = r8;
}

void ssaROP()
{
    uint64_t r8, a, b, c, d, e, f, g, h, i;

    a = 0x756AF83DE1FFE263;
    b = 0xFA34AE1002547B89;

    c = ~a;
    d = bswap(b);

    r8 = *(uint64_t *)&input[0];
    r8 ^= d;
    r8 += c;
    *(uint64_t *)&output[0] = r8;

    r8 = *(uint64_t *)&input[8];
    r8 ^= c;
    r8 = bswap(r8);
    r8 -= d;
    *(uint64_t *)&output[8] = r8;

    r8 = *(uint64_t *)&input[16];
    r8 = bswap(r8);
    e = rol(d, 13);
    r8 = ror(r8, e & 63);
    *(uint64_t *)&output[16] = r8;

    r8 = *(uint64_t *)&input[24];
    r8 ^= e;
    f = ror(c, 21);
    g = e;
    h = f;
    i = g;
    r8 = ror(r8, h & 63);
    r8 += i;
    *(uint64_t *)&output[24] = r8;
}

void unROP()
{
    uint64_t r8, a, b, c, d, e, f, g, h, i;

    a = 0x756AF83DE1FFE263;
    b = 0xFA34AE1002547B89;

    c = ~a;
    d = bswap(b);

    r8 = *(uint64_t *)&input[0];
    r8 -= c;
    r8 ^= d;
    *(uint64_t *)&output[0] = r8;

    r8 = *(uint64_t *)&input[8];
    r8 += d;
    r8 = bswap(r8);
    r8 ^= c;
    *(uint64_t *)&output[8] = r8;

    r8 = *(uint64_t *)&input[16];
    e = rol(d, 13);
    r8 = rol(r8, e & 63);
    r8 = bswap(r8);
    *(uint64_t *)&output[16] = r8;

    r8 = *(uint64_t *)&input[24];
    f = ror(c, 21);
    g = e;
    h = f;
    i = g;
    r8 -= i;
    r8 = rol(r8, h & 63);
    r8 ^= e;
    *(uint64_t *)&output[24] = r8;
}

// winning hash xored with mask

unsigned char winx[] =
{
  0x25, 0x67, 0x81, 0x29, 0x75, 0x6B, 0xE4, 0xD4, 0xE1, 0x7C, 
  0x9F, 0x34, 0x26, 0xE8, 0x1D, 0x33, 0x8E, 0x64, 0xE3, 0xC4, 
  0x66, 0x8D, 0x65, 0x7A, 0x20, 0xE8, 0x63, 0xE7, 0xBF, 0x1A, 
  0xDD, 0xE6,
};

void test(void)
{
    int i;

    encrypt();

    for (i = 0; i < 32; i++)
        printf("%02x", output[i]);
    printf("\n");

    memcpy(input, output, 32);

    ROP();

    for (i = 0; i < 32; i++)
        printf("%02x", output[i]);
    printf("\n");

    ssaROP();

    for (i = 0; i < 32; i++)
        printf("%02x", output[i]);
    printf("\n");

    memcpy(input, output, 32);

    unROP();

    for (i = 0; i < 32; i++)
        printf("%02x", output[i]);
    printf("\n");

    memcpy(input, output, 32);

    decrypt();

    for (i = 0; i < 32; i++)
        printf("%02x", output[i]);
    printf("\n");

    memcpy(input, output, 32);
}

void crack(void)
{
    int i;

    unROP();

    for (i = 0; i < 32; i++)
        printf("%02x", output[i]);
    printf("\n");

    memcpy(input, output, 32);

    decrypt();

    for (i = 0; i < 32; i++)
        printf("%02x", output[i]);
    printf("\n");
}

int main(int argc, char *argv[])
{
    if (argc > 1) {
        memcpy(input, argv[1], strlen(argv[1]));
        test();
    } else {
        memcpy(input, winx, 32);
        crack();
    }

    return 0;
}

