#include <stdio.h>
#include <stdint.h>

typedef unsigned char byte;

byte code[1024];
char disas[1024];

int main(int argc, char *argv[])
{
    int vip, b0, b1, opc, delta;
    uint32_t dw;
    FILE *in = fopen(argv[1], "rb");

    fread(code, 1, 1024, in);

    fclose(in);

    disas[0] = 1;
    //disas[0x2f9] = 1;
    //disas[0x308] = 1;

    for (vip = 0; vip < 1024;) {

        printf("%04x %02x ", vip, code[vip]);

        if (! disas[vip]) {
            printf("%02x\n", code[vip]);
            vip++;
            continue;
        }

        b0 = code[vip];
        vip++;
        opc = b0 & 0xf8;
        switch (opc) {
        case 0x00:
            printf("v2jmp   sword(R0)\n");
            break;
        case 0x18:
            b1 = code[vip];
            vip++;
            delta = ((b0 & 7) << 8) + b1;
            if (delta & 0x400)
                delta = -(delta & 0x3ff);
            printf("call    %04x\n", vip + delta);
            disas[vip+delta] = 1;
            break;
        case 0x28:
            printf("sse\n");
            break;
        case 0x30:
            dw = *(uint32_t *)&code[vip];
            vip += 4;
            delta = (dw >> 6) & 0x7ff;
            if (delta & 0x400)
                delta = -(delta & 0x3ff);
            if (b0 & 1) {
                printf("jeq     R%d, R%d, %04x\n", (dw >> 29) & 7, (dw >> 18) & 7, vip + delta);
            } else {
                printf("jeq     R%d, %x, %04x\n", (dw >> 29) & 7, (dw >> 19) & 0xff, vip + delta);
            }
            disas[vip+delta] = 1;
            break;
        case 0x48:
            printf("exitvm (trigger ROP chain)\n");
            goto no_disas_next;
            break;
        case 0x50:
            b1 = code[vip];
            vip++;
            printf("and     R%d, %x\n", b0 & 7, b1);
            break;
        case 0x58:
            printf("ret\n");
            goto no_disas_next;
            break;
        case 0x68:
            if (b0 & 2) {
                b1 = code[vip];
                vip++;
                printf("xor     R%d, R%d\n", (b1 >> 4) & 7, (b1 >> 1) & 7);
            } else {
                dw = *(uint32_t *)&code[vip];
                vip += 4;
                printf("xor     R%d, %x\n", dw >> 29, dw & 0x1fffffff);
            }
            break;
        case 0x78:
            printf("checksum stack\n");
            break;
        case 0x80:
            if (b0 & 2) {
                b1 = code[vip];
                vip++;
                printf("sub     R%d, R%d\n", (b1 >> 4) & 7, (b1 >> 1) & 7);
            } else {
                dw = *(uint32_t *)&code[vip];
                vip += 4;
                printf("sub     R%d, %x\n", dw >> 29, dw & 0x1fffffff);
            }
            break;
        case 0x90:
            printf("vm2fetch\n");
            break;
        case 0xa0:
            b1 = code[vip];
            vip++;
            delta = ((b0 & 7) << 8) + b1;
            if (delta & 0x400)
                delta = -(delta & 0x3ff);
            printf("jmp     %04x\n", vip + delta);
            disas[vip+delta] = 1;
            goto no_disas_next;
            break;
        case 0xa8:
            b1 = code[vip];
            vip++;
            printf("shr     R%d, byte(R%d)\n", (b0 & 7), (b1 >> 3) & 7);
            break;
        case 0xb0:
            if (b0 & 2) {
                b1 = code[vip];
                vip++;
                printf("add     R%d, R%d\n", (b1 >> 4) & 7, (b1 >> 1) & 7);
            } else {
                dw = *(uint32_t *)&code[vip];
                vip += 4;
                printf("add     R%d, %x\n", dw >> 29, dw & 0x1fffffff);
            }
            break;
        case 0xc8:
            if (b0 & 4) {
                b1 = code[vip];
                vip++;
                printf("push    R%d\n", (b1 >> 3) & 7);
            } else {
                dw = *(uint32_t *)&code[vip];
                vip += 4;
                printf("push    %8x\n", dw);
            }
            break;
        case 0xe0:
            b1 = code[vip];
            vip++;
            printf("shl     R%d, %d\n", b0 & 7, b1);
            break;
        case 0xe8:
            switch (b0 & 3) {
            case 0:
                dw = *(uint32_t *)&code[vip];
                vip += 4;
                printf("mov     R%d, %x\n", dw >> 29, dw & 0x1fffffff);
                break;
            case 1:
                b1 = code[vip];
                vip++;
                printf("mov     R%d, R%d\n", (b1 >> 4) & 7, (b1 >> 1) & 7);
                break;
            case 2:
                b1 = code[vip];
                vip++;
                printf("mov     [R%d], R%d\n", (b1 >> 4) & 7, (b1 >> 1) & 7);
                break;
            case 3:
                b1 = code[vip];
                vip++;
                printf("mov     R%d, [R%d]\n", (b1 >> 4) & 7, (b1 >> 1) & 7);                
                break;
            }
            break;
        case 0x08:
        case 0x38:
        case 0x60:
        case 0x88:
        case 0xd0:
        case 0xf0:
        case 0xf8:
            printf("exit\n");
            break;
        default:
            printf("???\n");
            goto done;
        }
        disas[vip] = 1;
no_disas_next:;
    }
done:;

    return 0;
}
